{*******************************************************}
{                                                       }
{       DBCalculatorCombo                               }
{       Version 1.0                                     }
{       Compiled 18.12.98                               }
{                                                       }
{       Copyright (c) 1998 by Jos Reyes Cedeo		}
{                                                       }
{       e-mail: jreyes_97@yahoo.com                     }
{       ICQ number: 13899790                            }
{                                                       }
{*******************************************************}

{
   This a combo style version of the CalculatorEdit. To know the
   functioning characteristics see CalculatorEdit.Pas. This version
   show a little window with a calculator when you press the arrow 
   button.
    

   PROPERTIES:
   
   property ButtonSeparation: Integer;  // Separation between buttons inside the calculator window
   property CalculatorBackColor: TColor; // Background color of the calculator window
   property CalculatorHeight: Integer; // Height of the calculator window.
                                       //  Note: the witdh of the calculator window is the parent editor width.
   property CalculatorButtonsColor: TColor; // Calculator buttons color, as the name indicated
   property CalculatorUpColor: TColor; // Used to drow the border of the button
   property CalculatorDownColor: TColor; // Used to drow the border of the button
   property BackFont: TFont; // Font of the Back, CE and C buttons
   property NumbersFont: TFont; // Font of the numbers buttons, the toggle sign button(-/+) 
                                // and the decimal separator button (.)
   property OperatorsFont: TFont; // Font of the '+', '-', '/', '*', '%', '1/x' and '=' buttons
   property OnDropDown: TNotifyEvent; // Triggered when a calculator window is showed


  This version (1.0) is for Delphi 3.x and Delphi 4.x.

  If you detect a problem or have any suggestion send me a message (jreyes_97@yahoo.com).

  FUTURE
  
  I going to release the dataware version soon.
}
unit CalculatorCombo;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Buttons, CalculatorEdit;

const
  CALCULATOR_STANDARD_HEIGHT = 100;
  BUTTON_SEPARATION_DEFAULT  = 6;

  // Back, CE and C strings
  STR_BACK  = 'BACK'; // Back
  STR_CE    = 'CE'; // CE
  STR_C     = 'C'; // C

  // Numbers chars
  STR_7 = '7';
  STR_8 = '8';
  STR_9 = '9';

  STR_4 = '4';
  STR_5 = '5';
  STR_6 = '6';

  STR_1 = '1';
  STR_2 = '2';
  STR_3 = '3';

  STR_0 = '0';
  STR_SIGN = '+/-';
  STR_SEP = '.';

  // Operators
  STR_DIV = '/';
  STR_SQRT = 'sqrt';

  STR_MULT = '*';
  STR_PERC = '%';

  STR_MINUS = '-';
  STR_INVERT = '1/x';

  STR_PLUS = '+';
  STR_EQUAL = '=';

type
  TBotonPresionado = ( bpNone,
                       bpBack, bpCE, bpC,
                       bpSiete, bpOcho, bpNueve,
                       bpCuatro, bpCinco, bpSeis,
                       bpUno, bpDos, bpTres,
                       bpCero, bpSigno, bpPunto,
                       bpDividir, bpRaiz,
                       bpMultiplicar, bpPorciento,
                       bpRestar, bpInverso,
                       bpSumar, bpIgual
                        );

  TCalculatorWindow = class;

  TCalculatorCombo = class(TCalculatorEdit)
  private
    FTextMargin: Integer;
    FCalculator: TCalculatorWindow;
    FButton: TSpeedButton;
    FBtnControl: TWinControl;
    FOnDropDown: TNotifyEvent;
    FCalculatorBackColor: TColor;
    FCalculatorButtonsColor,
    FCalculatorForeColor,
    FCalculatorUpColor,
    FCalculatorDownColor: TColor;
    FCalculatorHeight,
    FButtonSeparation: Integer;
    FBackFont,
    FNumbersFont,
    FOperatorsFont: TFont;
    function GetMinHeight: Integer;
    procedure SetEditRect;

    procedure SetBackFont( aValue: TFont );
    procedure SetNumbersFont( aValue: TFont );
    procedure SetOperatorsFont( aValue: TFont );

    procedure WMPaste(var Message: TMessage); message WM_PASTE;
    procedure WMCut(var Message: TMessage); message WM_CUT;
    procedure WMKillFocus(var Message: TWMKillFocus); message WM_KILLFOCUS;
    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMCancelMode(var Message: TCMCancelMode); message CM_CANCELMODE;
    procedure CMExit(var Message: TCMExit); message CM_EXIT;
    procedure CMEnter(var Message: TCMGotFocus); message CM_ENTER;
    procedure CMHintShow(var Message: TMessage); message CM_HINTSHOW;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure DropDown; dynamic;
    procedure CloseUp; dynamic;
  published
    property ButtonSeparation: Integer read FButtonSeparation write FButtonSeparation default BUTTON_SEPARATION_DEFAULT;
    property CalculatorBackColor: TColor read FCalculatorBackColor write FCalculatorBackColor default clNavy;
    property CalculatorHeight: Integer read FCalculatorHeight write FCalculatorHeight default CALCULATOR_STANDARD_HEIGHT;
    property CalculatorButtonsColor: TColor read FCalculatorButtonsColor write FCalculatorButtonsColor default clBtnFace;
    property CalculatorUpColor: TColor read FCalculatorUpColor write FCalculatorUpColor default clWhite;
    property CalculatorDownColor: TColor read FCalculatorDownColor write FCalculatorDownColor default clGray;
    property BackFont: TFont read FBackFont write SetBackFont;
    property NumbersFont: TFont read FNumbersFont write SetNumbersFont;
    property OperatorsFont: TFont read FOperatorsFont write SetOperatorsFont;
    property OnDropDown: TNotifyEvent read FOnDropDown write FOnDropDown;
  end;

{ TCalculatorWindow }

  TCalculatorWindow = class( TCustomControl )
  private
    FEdit: TCalculatorCombo;
    FButtonPressed,
    FButtonDown: TBotonPresionado;
    procedure DrawButton( aCanvas: TCanvas; B: TBotonPresionado;
                          Text: string );
    procedure ButtonRect( B: TBotonPresionado; var R: TRect );
    function ButtonToChar( B: TBotonPresionado ): char;
    function CharToButton( ch: char ): TBotonPresionado;
    procedure ButtonPushed( B: TBotonPresionado );
    procedure Draw( aCanvas: TCanvas );    
  protected
    procedure Paint; override;
    procedure CreateParams(var Params: TCreateParams); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer ); override;
    function MouseToButton( X, Y: Integer ): TBotonPresionado;
    procedure PushButton( B: TBotonPresionado );    
  public
    procedure Repintar;
    property Edit: TCalculatorCombo read FEdit;
  end;

{ TComboButton }

  TComboButton = class(TSpeedButton)
  protected
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
  end;  

procedure Register;

implementation

{ TCalculatorCombo }

constructor TCalculatorCombo.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  AutoSize := False;
  FBtnControl := TWinControl.Create(Self);
  FBtnControl.Width := 17;
  FBtnControl.Height := 17;
  FBtnControl.Visible := True;
  FBtnControl.Parent := Self;
  FButton := TComboButton.Create(Self);
  FButton.SetBounds(0, 0, FBtnControl.Width, FBtnControl.Height);
  FButton.Glyph.Handle := LoadBitmap(0, PChar(32738));
  FButton.Visible := True;
  FButton.Parent := FBtnControl;
  FCalculator := TCalculatorWindow.Create( Self );
  FCalculator.FEdit := Self;
  FCalculator.Parent := Self;
  FCalculator.Visible := False;
  FCalculatorBackColor := clNavy;
  FCalculatorHeight := CALCULATOR_STANDARD_HEIGHT;
  FCalculatorButtonsColor := clBtnFace;
  FCalculatorForeColor := clBlack;
  FButtonSeparation := BUTTON_SEPARATION_DEFAULT;
  FCalculatorUpColor := clWhite;
  FCalculatorDownColor := clGray;

  Height := 21;
  Width := 133;

  FBackFont := TFont.Create;
  FBackFont.Color := clMaroon;

  FNumbersFont := TFont.Create;
  FNumbersFont.Color := clRed;

  FOperatorsFont := TFont.Create;
  FOperatorsFont.Color := clBlue;
end;

destructor TCalculatorCombo.Destroy;
begin
  FBackFont.Free;
  FNumbersFont.Free;
  FOperatorsFont.Free;
  inherited Destroy;
end;

procedure TCalculatorCombo.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Key in [VK_UP, VK_DOWN, VK_NEXT, VK_PRIOR]) then
  begin
    if not FCalculator.Visible then DropDown;
    Key := 0;
  end;
end;

procedure TCalculatorCombo.KeyPress(var Key: Char);
var
  b: TBotonPresionado;
begin
  inherited KeyPress(Key);
  case Key of
    ^H, ^V, ^X, #32..#255:
        if (IsNumber( Key )or IsOperator( Key )or IsBackCEoC( Key ))and
           FCalculator.Visible then
        begin
          b := FCalculator.CharToButton( Key );
          FCalculator.PushButton( b );
        end;
    char(VK_RETURN): // Perform calculation
      begin
      end;
    char(VK_ESCAPE): // Perform reset
      begin
        if FCalculator.Visible then
             CloseUp;
        SelectAll;
        Key := #0;
      end;
  end;
end;


procedure TCalculatorCombo.WMPaste(var Message: TMessage);
begin
  inherited;
end;

procedure TCalculatorCombo.WMCut(var Message: TMessage);
begin
  inherited;
end;

procedure TCalculatorCombo.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN;
end;

procedure TCalculatorCombo.CreateWnd;
begin
  inherited CreateWnd;
  SetEditRect;
  FCalculator.HandleNeeded;
end;

procedure TCalculatorCombo.SetEditRect;
var
  Loc: TRect;
begin
  Loc.Bottom := ClientHeight + 1;  {+1 is workaround for windows paint bug}
  Loc.Right := FBtnControl.Left - 2;
  Loc.Top := 0;
  Loc.Left := 0;
  SendMessage(Handle, EM_SETRECTNP, 0, LongInt(@Loc));
end;

procedure TCalculatorCombo.WMSize(var Message: TWMSize);
var
  MinHeight: Integer;
begin
  inherited;
  if (csDesigning in ComponentState) then
    FCalculator.SetBounds(0, Height + 1, 10, 10);
  MinHeight := GetMinHeight;
  if Height < MinHeight then Height := MinHeight
  else begin
    if NewStyleControls then
      FBtnControl.SetBounds(ClientWidth - FButton.Width, 0, FButton.Width, ClientHeight)
    else
      FBtnControl.SetBounds(ClientWidth - FButton.Width, 1, FButton.Width, ClientHeight - 1);
    FButton.Height := FBtnControl.Height;
    SetEditRect;
  end;
end;

function TCalculatorCombo.GetMinHeight: Integer;
var
  DC: HDC;
  SaveFont: HFont;
  I: Integer;
  SysMetrics, Metrics: TTextMetric;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  I := SysMetrics.tmHeight;
  if I > Metrics.tmHeight then I := Metrics.tmHeight;
  FTextMargin := I div 4;
  Result := Metrics.tmHeight + FTextMargin + GetSystemMetrics(SM_CYBORDER) * 4 + 1;
end;

procedure TCalculatorCombo.CMFontChanged(var Message: TMessage);
begin
  inherited;
  GetMinHeight;
end;

procedure TCalculatorCombo.CMEnabledChanged(var Message: TMessage);
begin
  inherited;
  FButton.Enabled := Enabled;
end;

procedure TCalculatorCombo.WMKillFocus(var Message: TWMKillFocus);
var
  F: THandle;
begin
  inherited;
  F := GetFocus;
  if ( F <> FCalculator.Handle )and
     ( F <> Handle )and
     ( F <> FBtnControl.Handle ) then
    CloseUp;
end;

procedure TCalculatorCombo.CMCancelMode(var Message: TCMCancelMode);
begin
  with Message do
    if (Sender <> Self) and (Sender <> FBtnControl) and
      (Sender <> FButton) and (Sender <> FCalculator) then CloseUp;
end;

procedure TCalculatorCombo.CMHintShow(var Message: TMessage);
begin
  Message.Result := Integer(FCalculator.Visible);
end;

procedure TCalculatorCombo.DropDown;
var
  P: TPoint;
  Y: Integer;
  CalcWidth, CalcHeight: Integer;
  SysBorderWidth, SysBorderHeight: Integer;
begin
  if not FCalculator.Visible and (Width > 20) then
  begin
    if Assigned(FOnDropDown) then FOnDropDown(Self);
    SysBorderWidth := GetSystemMetrics(SM_CXBORDER);
    SysBorderHeight := GetSystemMetrics(SM_CYBORDER);
    P := ClientOrigin;
    if NewStyleControls then
    begin
      Dec(P.X, 2 * SysBorderWidth);
      Dec(P.Y, SysBorderHeight);
    end;
    CalcHeight := FCalculatorHeight;
    FCalculator.Height := CalcHeight;
    if NewStyleControls then
      Y := P.Y + ClientHeight + 3 * SysBorderHeight else
      Y := P.Y + Height - 1;
    if (Y + CalcHeight) > Screen.Height then
    begin
      Y := P.Y - CalcHeight + 1;
      if Y < 0 then
      begin
        if NewStyleControls then
          Y := P.Y + ClientHeight + 3 * SysBorderHeight else
          Y := P.Y + Height - 1;
      end;
    end;
    CalcWidth := Width;
    if CalcWidth = 0 then
    begin
      if NewStyleControls then
        CalcWidth := Width + 2 * SysBorderWidth else
        CalcWidth := Width - 4;
    end;
    if NewStyleControls then
      SetWindowPos(FCalculator.Handle, 0, P.X, Y, CalcWidth, CalcHeight, SWP_NOACTIVATE) else
      SetWindowPos (FCalculator.Handle, 0, P.X + Width - CalcWidth, Y, CalcWidth, CalcHeight, SWP_NOACTIVATE);

    FCalculator.Visible := True;
    Windows.SetFocus(Handle);
  end;
end;

procedure TCalculatorCombo.CloseUp;
begin
  FCalculator.Visible := False;
end;

procedure TCalculatorCombo.CMExit(var Message: TCMExit);
begin
  if FCalculator.Visible then DropDown;
end;

procedure TCalculatorCombo.CMEnter(var Message: TCMGotFocus);
begin
  if AutoSelect and not (csLButtonDown in ControlState) then SelectAll;
  inherited;
end;

procedure TCalculatorCombo.SetBackFont( aValue: TFont );
begin
  FBackFont.Assign( aValue );
end;

procedure TCalculatorCombo.SetNumbersFont( aValue: TFont );
begin
  FNumbersFont.Assign( aValue );
end;

procedure TCalculatorCombo.SetOperatorsFont( aValue: TFont );
begin
  FOperatorsFont.Assign( aValue );
end;

{ TCalculatorWindow }

procedure TCalculatorWindow.DrawButton( aCanvas: TCanvas; B: TBotonPresionado;
                                        Text: string );
var
  X, Y: Integer;
  R: TRect;
  aUp: boolean;
begin
  aUp := ( FButtonDown <> B );
  ButtonRect( B, R );

  aCanvas.FillRect( R );

  if aUp then
    aCanvas.Pen.Color := FEdit.CalculatorUpColor
  else
    aCanvas.Pen.Color := FEdit.CalculatorDownColor;
  aCanvas.MoveTo( R.left, R.bottom );
  aCanvas.LineTo( R.left, R.top );
  aCanvas.LineTo( R.right, R.top );

  if aUp then
    aCanvas.Pen.Color := FEdit.CalculatorDownColor
  else
    aCanvas.Pen.Color := FEdit.CalculatorUpColor;
  aCanvas.LineTo( R.right, R.bottom );
  aCanvas.LineTo( R.left, R.bottom );

  InflateRect( R, -1, -1 );
  X := R.left;
  if not aUp then X := X + 1;
  Y := R.top;
  if not aUp then Y := Y + 1;

  aCanvas.TextRect( R, X, Y, Text );
end;

procedure TCalculatorWindow.Draw( aCanvas: TCanvas );
var
  OldBrushColor: TColor;
  OldPenColor: TColor;
begin
  OldBrushColor := aCanvas.Brush.Color;
  OldPenColor := aCanvas.Pen.Color;
  try
    // Pintar background
    aCanvas.Brush.Color := FEdit.CalculatorBackColor;
    aCanvas.FillRect( ClientRect );
    aCanvas.Rectangle( ClientRect.left, ClientRect.top,
                      ClientRect.right, ClientRect.bottom );

    // Pintar botones BACK, CE y C
    aCanvas.Brush.Color := FEdit.CalculatorButtonsColor;

    aCanvas.Font := FEdit.BackFont;

    DrawButton( aCanvas, bpBack, STR_BACK );
    DrawButton( aCanvas, bpCE, STR_CE );
    DrawButton( aCanvas, bpC, STR_C );

    // Pintar botones de numeros, +/- y el punto decimal
    aCanvas.Font := FEdit.NumbersFont;

    DrawButton( aCanvas, bpSiete, STR_7 );
    DrawButton( aCanvas, bpOcho, STR_8 );
    DrawButton( aCanvas, bpNueve, STR_9 );

    DrawButton( aCanvas, bpCuatro, STR_4 );
    DrawButton( aCanvas, bpCinco, STR_5 );
    DrawButton( aCanvas, bpSeis, STR_6 );

    DrawButton( aCanvas, bpUno, STR_1 );
    DrawButton( aCanvas, bpDos, STR_2 );
    DrawButton( aCanvas, bpTres, STR_3 );

    DrawButton( aCanvas, bpCero, STR_0 );
    DrawButton( aCanvas, bpSigno, STR_SIGN );
    DrawButton( aCanvas, bpPunto, STR_SEP );

    // Pintar los operadores
    aCanvas.Font := FEdit.OperatorsFont;

    DrawButton( aCanvas, bpDividir, STR_DIV );
    DrawButton( aCanvas, bpRaiz, STR_SQRT );

    DrawButton( aCanvas, bpMultiplicar, STR_MULT );
    DrawButton( aCanvas, bpPorciento, STR_PERC );

    DrawButton( aCanvas, bpRestar, STR_MINUS );
    DrawButton( aCanvas, bpInverso, STR_INVERT );

    DrawButton( aCanvas, bpSumar, STR_PLUS );
    DrawButton( aCanvas, bpIgual, STR_EQUAL );
  finally
    aCanvas.Brush.Color := OldBrushColor;
    aCanvas.Pen.Color := OldPenColor;
  end;
end;

procedure TCalculatorWindow.Paint;
var
  bmp: TBitmap;
begin
  bmp := TBitmap.Create;
  try
    bmp.Width := Width;
    bmp.Height := Height;
    Draw( bmp.Canvas );
    Canvas.CopyRect( ClientRect, bmp.Canvas, ClientRect );
  finally
    bmp.free;
  end;
end;

procedure TCalculatorWindow.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or ES_MULTILINE or WS_POPUP;
end;

procedure TCalculatorWindow.ButtonRect( B: TBotonPresionado; var R: TRect );
          procedure GetRect( var R: TRect; XInit, XEnd, YInit, YEnd: Integer );
          var
            CincoX, CincoY: Integer;
          begin
            CincoX := (( ClientRect.right - ClientRect.left ) div 5 );
            CincoY := (( ClientRect.bottom - ClientRect.top ) div 5 );

            R.left := ClientRect.left + XInit*CincoX + ( FEdit.ButtonSeparation div 2 );
            R.right := ClientRect.left + XEnd*CincoX - ( FEdit.ButtonSeparation div 2 );
            R.top := ClientRect.top + YInit*CincoY + ( FEdit.ButtonSeparation div 2 );
            R.bottom := ClientRect.top + YEnd*CincoY - ( FEdit.ButtonSeparation div 2 );
          end;
begin
  case B of
    bpBack:        GetRect( R, 1, 3, 0, 1 );
    bpCE:          GetRect( R, 3, 4, 0, 1 );
    bpC:           GetRect( R, 4, 5, 0, 1 );

    bpSiete:       GetRect( R, 0, 1, 1, 2 );
    bpOcho:        GetRect( R, 1, 2, 1, 2 );
    bpNueve:       GetRect( R, 2, 3, 1, 2 );

    bpCuatro:      GetRect( R, 0, 1, 2, 3 );
    bpCinco:       GetRect( R, 1, 2, 2, 3 );
    bpSeis:        GetRect( R, 2, 3, 2, 3 );

    bpUno:         GetRect( R, 0, 1, 3, 4 );
    bpDos:         GetRect( R, 1, 2, 3, 4 );
    bpTres:        GetRect( R, 2, 3, 3, 4 );

    bpCero:        GetRect( R, 0, 1, 4, 5 );
    bpSigno:       GetRect( R, 1, 2, 4, 5 );
    bpPunto:       GetRect( R, 2, 3, 4, 5 );

    bpDividir:     GetRect( R, 3, 4, 1, 2 );
    bpRaiz:        GetRect( R, 4, 5, 1, 2 );

    bpMultiplicar: GetRect( R, 3, 4, 2, 3 );
    bpPorciento:   GetRect( R, 4, 5, 2, 3 );

    bpRestar:      GetRect( R, 3, 4, 3, 4 );
    bpInverso:     GetRect( R, 4, 5, 3, 4 );

    bpSumar:       GetRect( R, 3, 4, 4, 5 );
    bpIgual:       GetRect( R, 4, 5, 4, 5 );
  end; { case }
end;

function TCalculatorWindow.MouseToButton( X, Y: Integer ): TBotonPresionado;
var
  i: TBotonPresionado;
  R: TRect;
begin
  Result := bpNone;
  for i := bpBack to bpIgual do
  begin
    ButtonRect( i, R );
    if PtInRect( R, Point( X, Y )) then
    begin
      Result := i;
      exit;
    end;
  end;
end;

procedure TCalculatorWindow.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  i : Integer;
begin
  FEdit.SetFocus;
  if Button <> mbLeft then Exit;

  FButtonDown := MouseToButton( X, Y );
  if FButtonDown <> bpNone then
    FButtonPressed := FButtonDown
  else
   FButtonPressed := bpNone;
  Repintar;

  // Wait a few mili seconds
  for i := 0 to 200 do
    Application.ProcessMessages;
end;

procedure TCalculatorWindow.MouseMove( Shift: TShiftState; X, Y: Integer);
begin
  if FButtonPressed = bpNone then Exit;

  FButtonDown := MouseToButton( X, Y );
  if FButtonDown <> FButtonPressed then
    FButtonDown := bpNone;
  Repintar;
end;

function TCalculatorWindow.ButtonToChar( B: TBotonPresionado ): char;
begin
  case B of
    bpBack:        Result := CH_BACK;
    bpCE:          Result := CH_CE;
    bpC:           Result := CH_C;

    bpSiete:       Result := CH_7;
    bpOcho:        Result := CH_8;
    bpNueve:       Result := CH_9;

    bpCuatro:      Result := CH_4;
    bpCinco:       Result := CH_5;
    bpSeis:        Result := CH_6;

    bpUno:         Result := CH_1;
    bpDos:         Result := CH_2;
    bpTres:        Result := CH_3;

    bpCero:        Result := CH_0;
    bpSigno:       Result := CH_SIGN;
    bpPunto:       Result := DecimalSeparator;

    bpDividir:     Result := CH_DIV;
    bpRaiz:        Result := CH_SQRT;

    bpMultiplicar: Result := CH_MULT;
    bpPorciento:   Result := CH_PERC;

    bpRestar:      Result := CH_MINUS;
    bpInverso:     Result := CH_INVERT;

    bpSumar:       Result := CH_PLUS;
    bpIgual:       Result := CH_EQUAL;
  else Result := #0;
  end; { case }
end;

function TCalculatorWindow.CharToButton( ch: char ): TBotonPresionado;
begin
  case ch of
    CH_BACK,
    CH_BACK2,
    CH_BACK3: Result := bpBack;
    CH_CE,
    CH_CE2:    Result := bpCE;
    CH_C,
    CH_C2:    Result := bpC;

    CH_7:         Result := bpSiete;
    CH_8:         Result := bpOcho;
    CH_9:         Result := bpNueve;

    CH_4:         Result := bpCuatro;
    CH_5:         Result := bpCinco;
    CH_6:         Result := bpSeis;

    CH_1:         Result := bpUno;
    CH_2:         Result := bpDos;
    CH_3:         Result := bpTres;

    CH_0:         Result := bpCero;
    CH_SIGN,
    CH_SIGN2:    Result := bpSigno;
    CH_SEP,
    CH_SEP2:    Result := bpPunto;

    CH_DIV:         Result := bpDividir;
    CH_SQRT,
    CH_SQRT2:    Result := bpRaiz;

    CH_MULT:         Result := bpMultiplicar;
    CH_PERC:         Result := bpPorciento;

    CH_MINUS:         Result := bpRestar;
    CH_INVERT,
    CH_INVERT2:         Result := bpInverso;

    CH_PLUS:         Result := bpSumar;
    CH_EQUAL:         Result := bpIgual;
  else Result := bpNone;
  end; { case }
end;

procedure TCalculatorWindow.ButtonPushed( B: TBotonPresionado );
var
 ch: char;
begin
  ch := ButtonToChar( B );
  SendMessage( FEdit.Handle, WM_CHAR, word( ch ), 0 );
end;

procedure TCalculatorWindow.MouseUp( Button: TMouseButton; Shift: TShiftState;
                                     X, Y: Integer );
begin
  if FButtonPressed = bpNone then Exit;
  if Button <> mbLeft then Exit;

  FButtonDown := MouseToButton( X, Y );
  if FButtonDown = FButtonPressed then
    ButtonPushed( FButtonPressed );

  FButtonPressed := bpNone;
  FButtonDown := bpNone;
  Repintar;
end;

procedure TCalculatorWindow.Repintar;
begin
  InvalidateRect( Handle, nil, false );
end;

// Simulates an button being pushed
procedure TCalculatorWindow.PushButton( B: TBotonPresionado );
var
  OldButton: TBotonPresionado;
  i : Integer;
begin
  OldButton := FButtonDown;
  try
    FButtonDown := B;
    Repintar;
    // Wait a few mili seconds
    for i := 0 to 200 do
      Application.ProcessMessages;

    FButtonDown := OldButton;
    Repintar;    
  except
    FButtonDown := OldButton;
  end;
end;

{ TComboButton }

procedure TComboButton.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  with TCalculatorCombo( Parent.Parent ) do
    if not FCalculator.Visible then
      if (Handle <> GetFocus) and CanFocus then
      begin
        SetFocus;
        if GetFocus <> Handle then Exit;
      end;
  inherited MouseDown (Button, Shift, X, Y);
  with TCalculatorCombo( Parent.Parent ) do
    if FCalculator.Visible then CloseUp
    else DropDown;
end;

procedure TComboButton.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited MouseMove (Shift, X, Y);
  if (ssLeft in Shift) and (GetCapture = Parent.Handle) then
    begin
    end;
end;

procedure Register;
begin
  RegisterComponents('Jose', [TCalculatorCombo]);
end;

end.
