{*******************************************************}
{                                                       }
{       CalculatorEdit                                  }
{       Version 1.0                                     }
{       Compiled 18.12.98                               }
{                                                       }
{       Copyright (c) 1998 by Jos Reyes Cedeo		}
{                                                       }
{       e-mail: jreyes_97@yahoo.com                     }
{       ICQ number: 13899790                            }
{                                                       }
{*******************************************************}

{
   This a editor prepared to edit numbers. During the edition you
   can use the editor as a calculator, that is, you can type 10 + 10 =
   and the text inside the editor is replaced by the 20 character 
   string. The editor implement the following operations:
    '+' -> Addition
    '-' -> Substration
    '/' -> Division
    '*' -> Multiplication
    'r' -> Square root
    '%' -> Percent
    'x' -> 1 / x
    '=' -> Result of the calculation

   Also you can use the Backspace, the CE command ( using 'l' char ),
   and the C command ( using 'c' char ).
    

   PROPERTIES:
   
   property Value: Double read GetValue; // Returns the text inside the editor in
                                         //number format


  This version (1.0) is for Delphi 3.x and Delphi 4.x.

  If you detect a problem or have any suggestion send me a message (jreyes_97@yahoo.com).

  FUTURE
  
  I going to release the dataware version soon.
}
unit CalculatorEdit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls;

const
  // Back, CE and C characters
  CH_BACK  = #8; // Back valid chars
  CH_BACK2 = 'b';
  CH_BACK3 = 'B';

  CH_CE    = 'l'; // CE valid chars
  CH_CE2   = 'L';

  CH_C     = 'c'; // C valid chars
  CH_C2    = 'C';

  // Numbers chars
  CH_7 = '7';
  CH_8 = '8';
  CH_9 = '9';

  CH_4 = '4';
  CH_5 = '5';
  CH_6 = '6';

  CH_1 = '1';
  CH_2 = '2';
  CH_3 = '3';

  CH_0 = '0';
  CH_SIGN = 's';
  CH_SIGN2 = 'S';
  CH_SEP = '.';
  CH_SEP2 = ',';
  //  CH_SEPARATOR = DecimalSeparator; Can not define this character

  // Operators
  CH_DIV = '/';
  CH_SQRT = 'r';
  CH_SQRT2 = 'R';

  CH_MULT = '*';
  CH_PERC = '%';

  CH_MINUS = '-';
  CH_INVERT = 'x';
  CH_INVERT2 = 'X';

  CH_PLUS = '+';
  CH_EQUAL = '=';
type
  TOperator  = ( opNone,
                 opDivide, opSqrt,
                 opMultiplica, opPorciento,
                 opResta, opInverso,
                 opSuma, opIgual );

  TCalculatorEdit = class(TCustomEdit)
  private
    FAcumulate: Double;
    FLastOperator: TOperator;
    FClearOnNext: boolean; // Clear the text on the next number typed
    procedure ToggleSign;
    procedure ProcessOperation( op: TOperator );
    function GetValue: Double;
  protected
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure KeyPress(var Key: Char); override;
    procedure CreateParams(var Params: TCreateParams); override;
    function IsNumber( Key: char ): boolean;
    function IsOperator( Key: char ): boolean;
    function IsBackCEoC( var Key: char ): boolean;
    function CharToOperator( Key: char ): TOperator;
  public
    constructor Create( aOwner: TComponent ); override;
    property Value: Double read GetValue;
  published
    property AutoSelect;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ImeMode;
    property ImeName;
    property MaxLength;
    property ParentColor;
    property ParentCtl3D;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnChange;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

const
  BinaryOperators : set of TOperator = [ opDivide,
                                         opMultiplica,
                                         opResta,
                                         opSuma, opIgual,
                                         opPorciento ];

procedure Register;

implementation

{ TCalculatorEdit }

constructor TCalculatorEdit.Create( aOwner: TComponent );
begin
  inherited Create( aOwner );
  MaxLength := 20;
end;

procedure TCalculatorEdit.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
end;

function TCalculatorEdit.IsNumber( Key: char ): boolean;
begin
  Result := Key in [ CH_7, CH_8, CH_9,  // Number
                     CH_4, CH_5, CH_6,
                     CH_1, CH_2, CH_3,
                     CH_0,
                     CH_SIGN, CH_SIGN2, // Toggle sign
                     DecimalSeparator  ];
end;

function TCalculatorEdit.IsOperator( Key: char ): boolean;
begin
  Result := Key in [ CH_DIV, // Dividir
                     CH_SQRT, CH_SQRT2, // Raiz cuadrada
                     CH_MULT,  // Multiplicar
                     CH_PERC, // Porciento
                     CH_MINUS, // Sustraer
                     CH_INVERT, CH_INVERT2, // 1/x ( Inverso )
                     CH_PLUS,  // Adicionar
                     CH_EQUAL // Igual a
                   ];
end;

function TCalculatorEdit.IsBackCEoC( var Key: char ): boolean;
begin
  Result := Key in [ CH_BACK2, // Back
                     CH_BACK3,
                     CH_BACK,
                     CH_CE, // CE
                     CH_CE2,
                     CH_C, // C
                     CH_C2
                   ];
  if Key in [ CH_BACK2, CH_BACK3 ] then Key := CH_BACK;
end;

function TCalculatorEdit.CharToOperator( Key: char ): TOperator;
begin
  case Key of
      CH_DIV : Result := opDivide;
      CH_SQRT, CH_SQRT2 : Result := opSqrt;
      CH_MULT : Result := opMultiplica;
      CH_PERC : Result := opPorciento;
      CH_MINUS : Result := opResta;
      CH_INVERT, CH_INVERT2 : Result := opInverso;
      CH_PLUS : Result := opSuma;
      CH_EQUAL : Result := opIgual;
    else Result := opNone;
  end;
end;

procedure TCalculatorEdit.ProcessOperation( op: TOperator );
var
  Value : Double;
  flag: boolean;
begin
  try Value := StrToFloat( Text ); except Value := 0; end;

  if op in BinaryOperators then
  begin
    if FClearOnNext and ( op <> opIgual ) then
    begin
      FLastOperator := op;
    end
    else
    begin
      flag := true;
      if op = opPorciento then
      begin
        FAcumulate := ( FAcumulate * Value ) / 100;
      end
      else
      begin
        case FLastOperator of
          opSuma : FAcumulate := FAcumulate + Value;
          opDivide: FAcumulate := FAcumulate / Value;
          opMultiplica: FAcumulate := FAcumulate * Value;
          opResta: FAcumulate := FAcumulate - Value;
        else flag := false;
        end;
      end;
      FClearOnNext := true;
      if flag then
        Text := FloatToStr( FAcumulate )
      else FAcumulate := Value;
      if op <> opIgual then FLastOperator := op;
    end;
  end
  else
   begin
    case op of
      opSqrt : begin
                 FAcumulate := Sqrt( Value );
                 Text := FloatToStr( FAcumulate );
               end;
      opInverso: begin
                 FAcumulate := 1 / Value;
                 Text := FloatToStr( FAcumulate );
               end;
    end;
   end;
end;

function TCalculatorEdit.GetValue: Double;
begin
  Result := StrToFloat( Text );
end;

procedure TCalculatorEdit.ToggleSign;
begin
  if Text[ 1 ] = CH_MINUS then Text := Copy( Text, 2, Length( Text ))
  else Text := CH_MINUS + Text;
end;

procedure TCalculatorEdit.KeyPress(var Key: Char);
          procedure ProcessCEoC( Key: Char );
          begin
            case Key of
              CH_CE, // CE
              CH_CE2: begin
                     Text := CH_0;
                     SelectAll;
                   end;
              CH_C, // C
              CH_C2: begin
                     Text := CH_0;
                     FAcumulate := 0;
                     FLastOperator := opNone;
                     FClearOnNext := false;
                     SelectAll;
                   end;
            end;
          end;
var
  op: TOperator;
begin
  inherited KeyPress(Key);
  case Key of
    ^H, ^V, ^X, #32..#255:
      begin
        if not( IsNumber( Key )or IsOperator( Key )or IsBackCEoC( Key )) then
          Key := #0
        else
          begin
            if IsNumber( Key ) then
              begin
                if Key in [ CH_SIGN, CH_SIGN2 ] then
                begin
                  ToggleSign;
                  Key := #0;
                  Exit;
                end;
                if ( Key = DecimalSeparator )and
                   ( pos( DecimalSeparator, Text ) <> 0 )then
                begin
                  Key := #0;
                  Exit;
                end;
                if FClearOnNext then Text := '';
                FClearOnNext := false;
                Exit;
              end;
            if IsOperator( Key ) then
             begin
               op := CharToOperator( Key );
               ProcessOperation( op );
               Key := #0;
               Exit;
             end;
            if IsBackCEoC( Key ) then
              begin
                if Key <> CH_BACK then
                begin
                  ProcessCEoC( Key );
                  Key := #0;
                end;
                Exit;
              end;
          end;
      end;
    char(VK_RETURN): // Perform calculation
      begin
      end;
    char(VK_ESCAPE): // Perform reset
      begin
      end;
  end;
end;

procedure TCalculatorEdit.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or ES_MULTILINE or WS_CLIPCHILDREN or ES_RIGHT;
end;

procedure Register;
begin
  RegisterComponents('Jose', [TCalculatorEdit]);
end;

end.
